/*-------------------------------------------------------- 
 * Module Name : cattp and sms lib
 * Version : 1.7.1
 * 
 * Software Name : Transport Manager
 * Version : 1.7.1
 *
 * Copyright © 01-10-2012 – 01-10-2012 France Télécom
 * This software is distributed under the Apache (v2) license,
 * the text of which is available at http://www.apache.org/licenses/LICENSE-2.0.html
 * or see the "license.txt" file for more details.
 *--------------------------------------------------------
 * File Name   : ${file_name}
 *
 * Created     : ${date}
 * Author(s)   : ${authors}
 *
 * Description :
 * Main class of this lib. It shall be called by the upper level
 *
 *--------------------------------------------------------
 * ${Log}
 *
 */
package com.orange.labs.quickpass.tm;

import com.orange.labs.quickpass.enums.TransportBearer;
import com.orange.labs.quickpass.tm.cattp.server.CATTPServer;
import com.orange.labs.quickpass.tm.cattp.states.CATTPCloseWait;
import com.orange.labs.quickpass.tm.cattp.states.CATTPClosed;
import com.orange.labs.quickpass.tm.cattp.states.CATTPListen;
import com.orange.labs.quickpass.tm.cattp.states.CATTPOpen;
import com.orange.labs.quickpass.tm.cattp.states.CATTPState;
import com.orange.labs.quickpass.tm.cattp.states.CATTPSynRcvd;
import com.orange.labs.quickpass.tm.service.CattpSessionStatus;
import com.orange.labs.quickpass.tm.service.cattp.AlreadyConnectedException;
import com.orange.labs.quickpass.tm.service.cattp.CattpConnectionInfo;
import com.orange.labs.quickpass.tm.service.cattp.NoMorePortAvailableException;
import com.orange.labs.quickpass.tm.service.cattp.SessionNotFoundException;
import com.orange.labs.quickpass.tm.session.CATTPSession;
import com.orange.labs.quickpass.tm.session.SMSSession;
import com.orange.labs.quickpass.tm.session.SessionLocator;
import com.orange.labs.quickpass.tm.sms.gateway.SMSGateway;
import com.orange.labs.quickpass.tm.sms.server.SMSServer;
import com.orange.labs.quickpass.type.Msisdn;
import java.util.List;
import java.util.UUID;
import javax.annotation.Resource;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Implementation of the transport manager interface
 *
 */
public final class TransportManagerImpl implements TransportManager {

    private static final Logger LOGGER = LoggerFactory.getLogger(TransportManagerImpl.class);
    @Resource
    private CATTPServer cattpServer;
    @Resource
    private SMSServer smsServer;
    @Resource
    private SMSGateway smsGateway;
    @Resource
    private SessionLocator sessionLocator;

    private TransportManagerImpl() {
    }

    //{@inheritDoc} 
    @Override
    public CattpConnectionInfo openCattpLink(Msisdn msisdn) throws NoMorePortAvailableException, AlreadyConnectedException {
        if (!checkMsisdnAndCattpServer(msisdn)) {
            return null;
        }
        if (!cattpServer.getCattpServerThread().isAlive()) {

            LOGGER.warn(String.format(
                    "[%s]: CATTP server thread in state %s has to be restarted because of unexpected shutdown !",
                    msisdn, cattpServer.getCattpServerThread().getState()));
            cattpServer.getSock().close();
            cattpServer.setSock(null);
            cattpServer.start();
        }
        try {
            CattpSessionStatus status = getCattpSessionStatus(msisdn);
            if (status.equals(CattpSessionStatus.OPENED)) {
                throw new AlreadyConnectedException("Session already active for " + msisdn);
            }
        } catch (SessionNotFoundException sessionNotFoundException) {
            //normal behaviour
        }
        String ip = cattpServer.getPublicIpAddress();
        int udpPort = cattpServer.getPort();
        UUID sessionID = UUID.randomUUID();

        int cattpPort = cattpServer.createCattpSession(msisdn, sessionID.toString());
        return new CattpConnectionInfo(ip, udpPort, cattpPort, sessionID);
    }

    //{@inheritDoc} 
    @Override
    public void sendJavaPushSMS(Msisdn[] msisdns, String message, String port) {
        for (Msisdn oneMsisdn : msisdns) {
            SMSSession session = (SMSSession) sessionLocator.createTransportSession("smsSession");
            session.setMsisdn(oneMsisdn);
            session.setTextMsg(message);
            session.setPushPort(Integer.parseInt(port));
            // @FIXME : gestion des session pour le push également.
            // implémenter dans SmsSession
            //smsServer.executeSession(session);
            smsGateway.sendPushSMS(session);
        }
    }

    //{@inheritDoc} 
    @Override
    public void sendSMSText(Msisdn[] msisdns, String message) {
        for (Msisdn oneMsisdn : msisdns) {
            SMSSession session = (SMSSession) sessionLocator.createTransportSession("smsSession");
            session.setMsisdn(oneMsisdn);
            session.setTextMsg(message);
            smsGateway.sendText(session);
        }
    }

    //{@inheritDoc} 
    @Override
    public void sendWAPPushSMS(Msisdn[] msisdns, String message, String url) {
        for (Msisdn oneMsisdn : msisdns) {
            SMSSession session = (SMSSession) sessionLocator.createTransportSession("smsSession");
            session.setMsisdn(oneMsisdn);
            session.setTextMsg(url + ";" + message);
            session.setPushPort(2948);
            smsGateway.sendWapPushSMS(session);
        }
    }

    //{@inheritDoc} 
    @Override
    public void closeCattpLink(Msisdn msisdn) throws SessionNotFoundException {
        // Session Cattp
        CATTPSession session = cattpServer.getSessionByMsisdn(msisdn);
        if (session == null) {
            throw new SessionNotFoundException("Session not found for " + msisdn);
        }
        session.close();
    }

    //{@inheritDoc} 
    @Override
    public void sendSecurePackets(Msisdn msisdn, List<byte[]> securePackets,
            TransportBearer bearer) throws SessionNotFoundException {
        switch (bearer) {
            case SMS_TO_UICC:
                SMSSession sms = (SMSSession) sessionLocator.createTransportSession("smsSession");
                sms.setMsisdn(msisdn);
                sms.setSecurePackets(securePackets);
                sms.setSecurePacketCursor(0);
                smsServer.executeAndStoreSession(sms);
                break;
            case CATTP_OVER_UDP_TO_UICC:
                CATTPSession catSession = cattpServer.getSessionByMsisdn(msisdn);
                if (catSession == null) {
                    throw new SessionNotFoundException("Session not found for " + msisdn);
                }
                catSession.setSecurePackets(securePackets);
                continueIfAlreadyOpen(catSession);
                break;
            default:
                throw new UnsupportedOperationException("Bearer not supported for now !");
        }
    }

    private boolean checkMsisdnAndCattpServer(Msisdn msisdn) {
        boolean check = true;
        if (msisdn == null) {
            LOGGER.error(String.format("[%s]: No MSISDN was provided", msisdn));
            check = false;
        }
        if (!cattpServer.isRunning()) {
            LOGGER.error(String.format("[%s]: CATTP server is not running", msisdn));
            check = false;
        }
        return check;
    }

    private void continueIfAlreadyOpen(CATTPSession catSession) {
        CATTPState state = catSession.getState();
        if (state instanceof CATTPOpen) {
            // decrement because of at posteriori increment
            catSession.setSequenceNumber(catSession.getSequenceNumber() - 1);
            catSession.provisionPackets();
            catSession.setState(CATTPOpen.getInstance());
        } else if (state instanceof CATTPCloseWait || state instanceof CATTPClosed) {
            throw new SessionNotFoundException("Session being (or already) closed for " + catSession.getMsisdn());
        }
    }

    //{@inheritDoc} 
    @Override
    public CattpSessionStatus getCattpSessionStatus(Msisdn msisdn) throws SessionNotFoundException {

        CATTPSession catSession = cattpServer.getSessionByMsisdn(msisdn);
        if (catSession == null) {            
            throw new SessionNotFoundException("Session not found for " + msisdn);
        }
        if (catSession.getState().equals(CATTPOpen.getInstance())
                || catSession.getState().equals(CATTPSynRcvd.getInstance())) {
            return CattpSessionStatus.OPENED;
        } else if (catSession.getState().equals(CATTPClosed.getInstance())
                || catSession.getState().equals(CATTPListen.getInstance())) {
            // l'état de création est closed
            return CattpSessionStatus.CREATED;
        }
        return CattpSessionStatus.CLOSED;
    }
}
