/*-------------------------------------------------------- 
* Module Name : cattp and sms lib
* Version : 1.7.1
* 
* Software Name : Transport Manager
* Version : 1.7.1
*
* Copyright © 01-10-2012 – 01-10-2012 France Télécom
* This software is distributed under the Apache (v2) license,
* the text of which is available at http://www.apache.org/licenses/LICENSE-2.0.html
* or see the "license.txt" file for more details.
*
*--------------------------------------------------------
* File Name   : ${file_name}
*
* Created     : ${date}
* Author(s)   : ${authors}
*
* Description :
* Object representing a pdu header.
*
*--------------------------------------------------------
* ${Log}
*
*/

package com.orange.labs.quickpass.tm.cattp.pdu;

import com.orange.labs.quickpass.tm.cattp.server.CATTPServer;

/**
 * Object representing a pdu header.
 * CAT_TP header format
 * CAT_TP PDUs are based on this header. </BR>
 * S Y N A C K E AK RST NUL SEG Ver No. </BR>
 * 8-bit Header Length </BR>
 * N-bit Variable Header Area </BR>
 * 16-bit Source Port </BR>
 * 16-bit Destination Port </BR>
 * 16-bit Data Length </BR>
 * 16-bit Checksum </BR>
 * 16-bit Window Size </BR>
 * 16-bit Sequence Number </BR>
 * 16-bit Acknowledgment Number </BR>
 * N-bit Variable Header Area </BR>
 */
public class CATTPHeader {

    /**
     * RST_HEADER_LENGTH
     */
    public static final int RST_HEADER_LENGTH = 0x13;
    /**
     * ACK_HEADER_LENGTH
     */
    public static final int ACK_HEADER_LENGTH = 0x12;
    /**
     * SYNACK_HEADER_LENGTH
     */
    public static final int SYNACK_HEADER_LENGTH = 0x17 + CATTPServer.getInstance().getIdentity().length();
    
    private int headerLen;
    private int seq;
    private int ack;
    private int srcPort;
    private int dstPort;
    private int winSize;
    private Byte reasonCode;
    private SynExtraData synData;

    /**
     * @param seq sequence number
     * @param ack ack sequence number
     * @param srcPort source Port
     * @param dstPort destination Port
     * @param winSize windows size
     */
    public CATTPHeader(int seq, int ack, int srcPort, int dstPort, int winSize) {
        this.headerLen = ACK_HEADER_LENGTH;
        this.seq = seq;
        this.ack = ack;
        this.srcPort = srcPort;
        this.dstPort = dstPort;
        this.winSize = winSize;
    }

    /**
     * @param headerLen header length
     * @param seq sequence number
     * @param ack ack sequence number
     * @param srcPort source Port
     * @param dstPort destination Port
     * @param winSize windows size
     */
    public CATTPHeader(int headerLen, int seq, int ack, int srcPort, int dstPort, int winSize) {
        this.headerLen = headerLen;
        this.seq = seq;
        this.ack = ack;
        this.srcPort = srcPort;
        this.dstPort = dstPort;
        this.winSize = winSize;
    }

    /**
     * @param seq sequence number
     * @param ack ack sequence number
     * @param srcPort source Port
     * @param dstPort destination Port
     * @param winSize windows size
     * @param reasonCode reason code
     */
    public CATTPHeader(int seq, int ack, int srcPort, int dstPort, int winSize, byte reasonCode) {
        this.headerLen = RST_HEADER_LENGTH;
        this.seq = seq;
        this.ack = ack;
        this.srcPort = srcPort;
        this.dstPort = dstPort;
        this.winSize = winSize;
        this.reasonCode = reasonCode;
    }

    /**
     * @param headerLen header length
     * @param seq sequence number
     * @param ack ack sequence number
     * @param srcPort source Port
     * @param dstPort destination Port
     * @param winSize windows size
     * @param reasonCode reason code
     */
    public CATTPHeader(int headerLen, int seq, int ack, int srcPort, int dstPort, int winSize, byte reasonCode) {
        this.headerLen = headerLen;
        this.seq = seq;
        this.ack = ack;
        this.srcPort = srcPort;
        this.dstPort = dstPort;
        this.winSize = winSize;
        this.reasonCode = reasonCode;
    }

    /**
     * this constructor shall be used only for syn pdu
     * @param seq sequence number
     * @param ack ack sequence number
     * @param srcPort source Port
     * @param dstPort destination Port
     * @param winSize windows size
     * @param extraData syn extra data
     */
    public CATTPHeader(int seq, int ack, int srcPort, int dstPort, int winSize, SynExtraData extraData) {
        this.headerLen = SYNACK_HEADER_LENGTH;
        this.seq = seq;
        this.ack = ack;
        this.srcPort = srcPort;
        this.dstPort = dstPort;
        this.winSize = winSize;
        this.synData = extraData;
    }

    /**
     * this constructor shall be used only for syn pdu
     * @param headerLen header length
     * @param seq sequence number
     * @param ack ack sequence number
     * @param srcPort source Port
     * @param dstPort destination Port
     * @param winSize windows size
     * @param extraData syn extra data
     */
    public CATTPHeader(int headerLen, int seq, int ack, int srcPort, int dstPort, int winSize, SynExtraData extraData) {
        this.headerLen = headerLen;
        this.seq = seq;
        this.ack = ack;
        this.srcPort = srcPort;
        this.dstPort = dstPort;
        this.winSize = winSize;
        this.synData = extraData;
    }

    /**
     *
     * @return the ack
     */
    public int getAck() {
        return ack;
    }

    /**
     *
     * @return destination port
     */
    public int getDstPort() {
        return dstPort;
    }

    /**
     *
     * @return lenght of header
     */
    public int getHeaderLen() {
        return headerLen;
    }

    /**
     *
     * @return the data Synchronisation
     */
    public SynExtraData getSynData() {
        return synData;
    }

    /**
     *
     * @return the reason code
     */
    public Byte getReasonCode() {
        return reasonCode;
    }

    /**
     *
     * @return the sequence
     */
    public int getSeq() {
        return seq;
    }

    /**
     *
     * @return the source port
     */
    public int getSrcPort() {
        return srcPort;
    }

    /**
     *
     * @return the windows size
     */
    public int getWinSize() {
        return winSize;
    }

    /**
     *
     * @param ack  the acknoledge
     */
    public void setAck(int ack) {
        this.ack = ack;
    }

    /**
     *
     * @param seq the sequence number
     */
    public void setSeq(int seq) {
        this.seq = seq;
    }

}
