/*-------------------------------------------------------- 
* Module Name : cattp and sms lib
* Version : 1.7.1
* 
* Software Name : Transport Manager
* Version : 1.7.1
*
* Copyright © 01-10-2012 – 01-10-2012 France Télécom
* This software is distributed under the Apache (v2) license,
* the text of which is available at http://www.apache.org/licenses/LICENSE-2.0.html
* or see the "license.txt" file for more details.
*
*--------------------------------------------------------
* File Name   : ${file_name}
*
* Created     : ${date}
* Author(s)   : ${authors}
*
* Description :
* Object representing a cattp pdu.
* 
* It can be used to parse datagram or to build a cattp pdu from scratch
*
*--------------------------------------------------------
* ${Log}
*
*/

package com.orange.labs.quickpass.tm.cattp.pdu;

import com.orange.labs.quickpass.common.utils.ConvertUtils;
import com.orange.labs.quickpass.tm.cattp.exceptions.BadCATTPChecksumException;
import com.orange.labs.quickpass.tm.cattp.server.CATTPServer;
import com.orange.labs.quickpass.tm.cattp.utils.CATTPUtils;
import com.orange.labs.quickpass.tm.session.CATTPSession;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A CATTP PDU see ETSI 102 127 
 * </p>
 * It can be used to parse datagram or to build a cattp pdu from scratch
 * @author Gael Gerard and Sylvain Decourval
 *
 */
public class CATTPPdu implements Comparable<CATTPPdu> {

    /**
     * SYN
     */
    public static final byte SYN = (byte) 0x80;
    /**
     * ACK
     */
    public static final byte ACK = (byte) 0x40;
    /**
     * EAK
     */
    public static final byte EAK = (byte) 0x20;
    /**
     * RST
     */
    public static final byte RST = (byte) 0x10;
    /**
     * NUL
     */
    public static final byte NUL = (byte) 0x08;
    /**
     * SEG
     */
    public static final byte SEG = (byte) 0x04;
    /**
     * LOGGER
     */
    protected static final Logger LOGGER = LoggerFactory.getLogger(CATTPPdu.class);
    /**
     * retry action which is scheduled for this pdu
     */
    @SuppressWarnings("rawtypes")
    protected ScheduledFuture futureRetry;
    /**
     * retryIdx
     */
    protected int retryIdx = -1;
    /**
     * header
     */
    protected CATTPHeader header;
    /**
     * resetCause
     */
    protected byte resetCause = 0;
    /**
     * dataLen
     */
    protected int dataLen = 0;
    /**
     * UDP datagram that built this CATTP PDU
     */
    protected byte[] datagram = null;
    /**
     * Data found when this Cattp pdu is ACKed
     */
    protected byte[] data = null;
    /**
     * maxPduSize
     */
    protected short maxPduSize;
    /**
     * maxSduSize
     */
    protected short maxSduSize;
    /**
     * ident
     */
    protected byte[] ident = null;
    /**
     * nbOutOfSequencePdusFound
     */
    protected int nbOutOfSequencePdusFound = 0;
    /**
     * outOfSequencePduArray
     */
    protected ArrayList<Short> outOfSequencePduArray = new ArrayList<Short>();
    /**
     * Associated session
     */
    protected CATTPSession session = null;

    /**
     * @param datagramData datagramData
     */
    public CATTPPdu(byte[] datagramData) {
        this(datagramData, null);
    }

    /**
     * Constructor from udp datagram (parsing datagram)
     * @param datagramData datagramData
     * @param session session (may be null)
     */
    public CATTPPdu(byte[] datagramData, CATTPSession session) {
        if (datagramData != null) {
            this.datagram = datagramData.clone();
            this.header = new CATTPHeader(this.datagram[3] & 0xff,
                    (this.datagram[10] << 8 | this.datagram[11] & 0xff) & 0xffff,
                    (this.datagram[12] << 8 | this.datagram[13] & 0xff) & 0xffff,
                    (this.datagram[4] << 8 | this.datagram[5] & 0xff) & 0xffff,
                    (this.datagram[6] << 8 | this.datagram[7] & 0xff) & 0xffff,
                    (this.datagram[14] << 8 | this.datagram[15] & 0xff) & 0xffff);
            this.dataLen = (this.datagram[8] << 8 | this.datagram[9] & 0xff) & 0xffff;

            if (hasData()) {
                data = new byte[this.datagram.length - 18];
                System.arraycopy(this.datagram, 18, data, 0, this.datagram.length - 18);
            }

            if (isSYNSet()) {
                maxPduSize = (short) ((this.datagram[18] << 8 | this.datagram[19] & 0xff) & 0xffff);
                maxSduSize = (short) ((this.datagram[20] << 8 | this.datagram[21] & 0xff) & 0xffff);
                ident = new byte[this.datagram[22]];
                System.arraycopy(this.datagram, 23, ident, 0, ident.length);
            }

            if (isRSTSet()) {
                resetCause = this.datagram[18];
            }

            if (isEAKSet()) {
                nbOutOfSequencePdusFound = (header.getHeaderLen() - 0x12) / 2;
                for (int i = 0; i < nbOutOfSequencePdusFound; i++) {
                    short outOfSequencePdu = (short) ((this.datagram[18 + i * 2] << 8 | this.datagram[19 + i * 2]) & 0xffff);
                    LOGGER.debug(String.format("%s Found PDU %s was out of sequence!", session, outOfSequencePdu));
                    outOfSequencePduArray.add(outOfSequencePdu);
                }
            }
        }

        this.session = session;
    }

    /**
     * Constructor (without building datagram). To build dtagram, please use
     * inherited constructor
     * @param header header
     * @param securePacket securePacket data
     * @param segmented is the pdu segmented ?
     * @param session current cattp session
     * @throws BadCATTPChecksumException
     */
    protected  CATTPPdu(CATTPHeader header, byte[] securePacket, boolean segmented, CATTPSession session)
            throws BadCATTPChecksumException {
        this.data = securePacket;
        this.header = header;
        this.datagram = null;
        if (data != null) {
            this.dataLen = securePacket.length;
        } else {
            dataLen = 0;
        }
        this.session = session;
    }

    /**
     * @param pduType pduType
     * @param header header
     * @param securePacket securePacket
     * @param segmented segmented
     * @return the pdu
     */
    protected static byte[] buildPdu(byte pduType, CATTPHeader header, byte[] securePacket, boolean segmented) {
        int pduDataLen = 0;
        if (securePacket != null) {
            pduDataLen = securePacket.length;
        }

        ByteBuffer buf = ByteBuffer.allocate(pduDataLen + header.getHeaderLen());
        if (segmented) {
            buf.put((byte) (pduType | CATTPPdu.SEG));
        } else {
            buf.put(pduType);
        }
        buf.putShort((short) 0);
        buf.put((byte) header.getHeaderLen());
        buf.position(4);
        buf.putShort((short) header.getSrcPort());
        buf.putShort((short) header.getDstPort());
        buf.position(8);
        if (securePacket == null) {
            buf.putShort((short) 0);
        } else {
            // correction du CPI non positionn� par le KMS
            buf.putShort((short) securePacket.length);
            buf.position(18);
            buf.put(securePacket);
        }
        buf.position(10);
        buf.putShort((short) header.getSeq());
        buf.putShort((short) header.getAck());
        buf.putShort((short) header.getWinSize());

        if (header.getReasonCode() != null) {
            buf.position(18);
            buf.put(header.getReasonCode());
        }

        if (header.getSynData() != null) {
            buf.position(18);
            buf.putShort((short) header.getSynData().getMaxPduSize());
            buf.putShort((short) header.getSynData().getMaxSduSize());
            buf.put((byte) header.getSynData().getIdentificationData().length);
            buf.put(header.getSynData().getIdentificationData());
        }

        try {
            CATTPUtils.checksum(buf.array(), false);
        } catch (BadCATTPChecksumException checksumException) {
            LOGGER.warn("Bad Checksum !!", checksumException);
        }

        return buf.array();
    }

    /**
     * @return true if true, false otherwise
     */
    public boolean verifyChecksum() {
        try {
            CATTPUtils.checksum(datagram, true);
        } catch (BadCATTPChecksumException e) {
            return false;
        }
        return true;
    }

    /**
     * Dont use with received PDU
     */
    public void send() {        
        if (session.getReceivedAckablePdu() != null) {
            setAck(session.getReceivedAckablePdu().getHeader().getSeq());
        }
        LOGGER.info(String.format("<<---[ SERVER ]-- %s </BR>for %s", this, session));
        CATTPServer.getInstance().sendDataGram(session.getIpAddress(), session.getUdpPort(), datagram);
    }

    /**
     * Should start sending the sentPdu, and trigger the retry mechanism
     *
     * @return the scheduler
     */
    @SuppressWarnings("rawtypes")
    public ScheduledFuture sendAndScheduleNextRetry() {
        if (this != null) {
            int delay = CATTPServer.getInstance().getAckPduRetransmissionTimerValue();
            if (isNULSet()) {
                delay = CATTPServer.getInstance().getNulPduRetransmissionTimerValue();
            }
            if (isRSTSet()) {
                delay = 3;
            }
            int nbRetries = CATTPServer.getInstance().getRetriesCounter();
            if (CATTPServer.getInstance().getTaskExecutor() != null) {
                futureRetry = ((ScheduledThreadPoolExecutor) CATTPServer.getInstance().getTaskExecutor().
                        getConcurrentExecutor()).schedule(new RetryPduTask(this, ++retryIdx), delay, TimeUnit.SECONDS);
                LOGGER.trace(
                        String.format("%s Retry timer set to %s seconds with %s retries for pdu %s with intial retry counter %s",
                        session, delay, nbRetries, getHeader().getSeq(), retryIdx));
                send();
                return futureRetry;
            } else {
                LOGGER.warn(String.format("%s Retry scheduler not set ! No retry set for the PDU %s", session, this));
                send();
            }
        }
        return null;
    }

    @Override
    /**
     * Used to display pdu contents
     */
    public String toString() {
        String res = "";
        if (datagram != null) {
            res += "[DST = " + Integer.toHexString(header.getDstPort()) + "]";
            res += "-[SEQ = " + header.getSeq() + "]";
            res += "-[ACK = " + header.getAck() + "/win = " + header.getWinSize() + "]";
            res += "Header :";
            if (isSYNSet()) {
                res += "-[SYN]";
                res += ", id : " + ConvertUtils.toHexString(this.datagram, " ", 22, header.getHeaderLen() - 22);
            }
            if (isACKSet()) {
                res += "-[ACK]";
            }
            if (isEAKSet()) {
                res += "-[EAK]";
                for (Short seq : outOfSequencePduArray) {
                    if (seq != null) {
                        res += "-[OUT OF SEQ PDU " + seq + "]";
                    }
                }
            }
            if (isRSTSet()) {
                res += "-[RST cause = " + resetCause + " ]";
            }
            if (isNULSet()) {
                res += "-[NUL]";
            }
            if (isSEGSet()) {
                res += "-[SEG]";
            }
            if (hasData()) {
                res += "-[DATALEN = '" + data.length + "']";
                res += "-[DATA = '" + ConvertUtils.toHexString(data, ":", 40) + "']";
            } else {
                res += "-[no DATA]";
            }
        }

        if (res.length() > 400) {
            res = res.substring(0, 300) + " .... " + res.substring(res.length() - 100, res.length());
        }

        return res;

    }

    /**
     *
     * @return true if synch is set
     */
    public boolean isSYNSet() {
        return (this.datagram[0] & SYN) == SYN;
    }

    /**
     *
     * @return true if acknoledge is set
     */
    public boolean isACKSet() {
        return (this.datagram[0] & ACK) == ACK;
    }

    /**
     *
     * @return true if reset is set
     */
    public boolean isRSTSet() {
        return (this.datagram[0] & RST) == RST;
    }

    /**
     *
     * @return true if ack null is set
     */
    public boolean isNULSet() {
        return (this.datagram[0] & NUL) == NUL;
    }

    /**
     *
     * @return true if eack null is set
     */
    public boolean isEAKSet() {
        return (this.datagram[0] & EAK) == EAK;
    }

    /**
     *
     * @return true if segment null is set
     */
    public boolean isSEGSet() {
        return (this.datagram[0] & SEG) == SEG;
    }

    /**
     *
     * @return the lenght of data
     */
    public int getDataLen() {
        return dataLen;
    }

    /**
     * @return true or false
     */
    public boolean hasData() {
        return dataLen != 0;
    }

    /**
     * @return true or false
     */
    public boolean needsAck() {
        return (hasData() || isSYNSet() || isNULSet());
    }

    /**
     *
     * @return true or false depending if PDU follows correctly the CATTP sequence
     */
    public boolean isInSequence() {
        return (session.getLastInSequencePdu() == null)
                || header.getSeq() == session.getLastInSequencePdu().getHeader().getSeq()
                || header.getSeq() == session.getLastInSequencePdu().getHeader().getSeq() + 1;
    }

    /**
     *
     * @return the pdu data
     */
    public byte[] getData() {
        return data.clone();
    }

    /**
     *
     * @param data the data
     */
    public void setData(byte[] data) {
        this.data = data.clone();
    }

    /**
     * @param seq seq
     */
    public void setSeq(int seq) {
        header.setSeq(seq);
        this.datagram[10] = (byte) ((seq >> 8) & 0xff);
        this.datagram[11] = (byte) (seq & 0xff);
    }

    /**
     * @param ack ack
     */
    public void setAck(int ack) {
        header.setAck(ack);
        this.datagram[12] = (byte) ((ack >> 8) & 0xff);
        this.datagram[13] = (byte) (ack & 0xff);
    }

    /**
     *
     * @param dataLen the lenght of data
     */
    public void setDataLen(int dataLen) {
        this.dataLen = dataLen;
    }

    /**
     *
     * @return the max pdu size
     */
    public short getMaxPduSize() {
        return maxPduSize;
    }

    /**
     *
     * @return the max sdu size
     */
    public short getMaxSduSize() {
        return maxSduSize;
    }

    /**
     *
     * @return the ident
     */
    public byte[] getIdent() {
        return ident.clone();
    }

    /**
     *
     * @return the datagram
     */
    public byte[] getDatagram() {
        return datagram;
    }

    /**
     *
     * @return the header
     */
    public CATTPHeader getHeader() {
        return header;
    }

    /**
     * Sorted according to seq number
     *
     * @param pdu oject
     * @return 1, -1 or 0
     */
    @Override
    public int compareTo(CATTPPdu pdu) {
        if (header.getSeq() > pdu.getHeader().getSeq()) {
            return 1;
        } else if (header.getSeq() < ((CATTPPdu) pdu).getHeader().getSeq()) {
            return -1;
        }
        return 0;
    }

    /**
     * 
     * @return out of sequence pdu array
     */
    public ArrayList<Short> getOutOfSequencePduArray() {
        return outOfSequencePduArray;
    }

    /**
     * cancel retries
     */
    public void cancelRetry() {
        int seq = getHeader().getSeq();
        if (futureRetry != null) {
            if (!futureRetry.isCancelled()) {
                futureRetry.cancel(true);
                ((ScheduledThreadPoolExecutor) CATTPServer.getInstance().getTaskExecutor().getConcurrentExecutor()).purge();
                LOGGER.debug(String.format("%s Cancel Retry for pdu %s which was scheduled in %s sec.",
                        session, seq, futureRetry.getDelay(TimeUnit.SECONDS)));
                futureRetry = null;
            } else {
                LOGGER.trace(String.format("%s Cancel retry action already done for pdu %s", session, seq));
            }
        } else {
            LOGGER.trace(String.format("%s Try to cancel a retry action which are not scheduled for pdu %s", session, seq));
        }

    }

    /**
     *
     * @return the cattp session
     */
    public CATTPSession getSession() {
        return session;
    }

    /**
     *
     * @param session  the cattp session
     */
    public void setSession(CATTPSession session) {
        this.session = session;
    }

    /**
     *
     * @return retry
     */
    @SuppressWarnings("rawtypes")
    public ScheduledFuture getFutureRetry() {
        return futureRetry;
    }
}
