/*-------------------------------------------------------- 
 * Module Name : cattp and sms lib
 * Version : 1.7.1
 * 
 * Software Name : Transport Manager
 * Version : 1.7.1
 *
 * Copyright © 01-10-2012 – 01-10-2012 France Télécom
 * This software is distributed under the Apache (v2) license,
 * the text of which is available at http://www.apache.org/licenses/LICENSE-2.0.html
 * or see the "license.txt" file for more details.
 *
 *--------------------------------------------------------
 * File Name   : ${file_name}
 *
 * Created     : ${date}
 * Author(s)   : ${authors}
 *
 * Description :
 * CATTP Server based on udp server (send and receive datagram)
 *
 *--------------------------------------------------------
 * ${Log}
 *
 */
package com.orange.labs.quickpass.tm.cattp.server;

import com.orange.labs.quickpass.tm.cattp.pdu.CATTPHeader;
import com.orange.labs.quickpass.tm.cattp.pdu.CATTPPdu;
import com.orange.labs.quickpass.tm.cattp.pdu.RstPdu;
import com.orange.labs.quickpass.tm.cattp.states.CATTPClosed;
import com.orange.labs.quickpass.tm.service.cattp.NoMorePortAvailableException;
import com.orange.labs.quickpass.tm.service.cattp.SessionNotFoundException;
import com.orange.labs.quickpass.tm.session.CATTPSession;
import com.orange.labs.quickpass.tm.session.SessionLocator;
import com.orange.labs.quickpass.type.Msisdn;
import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.MalformedURLException;
import java.net.SocketException;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import javax.annotation.PostConstruct;
import javax.annotation.Resource;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.scheduling.concurrent.ConcurrentTaskExecutor;

/**
 * The CATTP server, handles CATTP PDUs and sessions
 */
public class CATTPServer extends AbstractUDPServer {

    private static final Logger LOGGER = LoggerFactory.getLogger(CATTPServer.class);
    private static CATTPServer instance = null;
    /**
     * Sessions stored by CATTP port
     */
    private static Map<Integer, Runnable> sessionsByCattpPort = Collections.synchronizedMap(new HashMap<Integer, Runnable>());
    /**
     * Sessions stored by msisdn
     */
    private static Map<String, Runnable> sessionsByMsisdn = Collections.synchronizedMap(new HashMap<String, Runnable>());
    /**
     * Server CATTP identification data should be modified
     */
    /**
     * The first available port
     */
    private static final int FIRST_AVAILABALE_PORT = 0x401;
    /**
     * The last available port
     */
    private static final int MAX_AVAILABALE_PORT = 0xffff;
    /**
     * cattpport
     */
    public int cattpport = FIRST_AVAILABALE_PORT;
    @Value("${serverName}")
    private String identity = "Orange Labs CATTP server";
    /**
     * Number of CATTP PDUs retries
     */
    @Value("${retriesCounter}")
    private int retriesCounter = 3;
    /**
     * Retransmission Timer value in seconds
     */
    @Value("${ackPduRetransmissionTimerValue}")
    private int ackPduRetransmissionTimerValue = 10;
    /**
     * Retransmission Timer value in seconds
     */
    @Value("${nulPduRetransmissionTimerValue}")
    private int nulPduRetransmissionTimerValue = 5;
    /**
     * Time to live for an Idle Session
     */
    @Value("${sessionIdleLifeTime}")
    private int sessionIdleLifeTime = 60;
    @Resource
    private ConcurrentTaskExecutor taskExecutor;
    @Resource
    private SessionLocator sessionLocator;
    /**
     * The last received pdu for this session
     */
    private CATTPPdu receivedPdu;
    private Thread cattpServerThread;

    /**
     *
     * @return instance of cattpserver
     */
    public static CATTPServer getInstance() {
        if (instance == null) {
            instance = new CATTPServer();
        }
        return instance;
    }

    /**
     * returns a valid cattp port
     *
     * @param msisdn targeted msisdn
     * @param sessionId the sessionId
     * @return an available cattp port
     * @throws NoMorePortAvailableException
     */
    public int createCattpSession(Msisdn msisdn, String sessionId) throws NoMorePortAvailableException {
        if (msisdn == null) {
            return 0;
        }
        cattpport = locateAvailablePort();
        LOGGER.info(String.format("[%s]: Create new cattp session %s ", msisdn, sessionId));
        CATTPSession session = createNewSession(cattpport, msisdn, sessionId);
        linkSessionToMsisdn(msisdn, session);
        //démarrage du thread et mise en attente
        taskExecutor.execute(session);
        int queueSize = ((ScheduledThreadPoolExecutor) (taskExecutor.getConcurrentExecutor())).getQueue().size();
        int activeThreadNb = ((ScheduledThreadPoolExecutor) (taskExecutor.getConcurrentExecutor())).getActiveCount();
        LOGGER.info(String.format("[%s]: There are %s active threads and %s in the queue.",
                msisdn, activeThreadNb, queueSize));
        return cattpport;
    }

    /**
     * to be implemented
     */
    public void close() {
    }

    /**
     * removes a session from tables
     *
     * @param session session
     */
    public static void removeSession(CATTPSession session) {
        Msisdn msisdn = session.getMsisdn();
        CATTPSession currentSession = (CATTPSession) sessionsByMsisdn.get(msisdn);
        if (currentSession.getState() instanceof CATTPClosed) {
            sessionsByCattpPort.remove(Integer.valueOf(session.getCattpPort()));
            sessionsByMsisdn.remove(session.getMsisdn().getValue());
            LOGGER.info(String.format("%s session removed!", session));
        } else {
            LOGGER.info(String.format("The session %s has been replaced by %s", session, currentSession));
        }
    }

    /**
     * This server runner method
     */
    @Override
    public void run() {
        if (getSock() == null) {
            try {
                setSock(new DatagramSocket(port));
                LOGGER.debug("UDP server listening on port : " + port);

            } catch (SocketException e) {
                LOGGER.error(e.getMessage());
            }
        }

        LOGGER.debug("UDP Server Waiting...");
        running = true;
        while (running) {
            DatagramPacket indatagram = new DatagramPacket(inBuffer, inBuffer.length);
            try {
                getSock().receive(indatagram);
            } catch (IOException e) {
                running = false;
                LOGGER.error(e.getMessage());
                return;
            } catch (Exception e) {
                LOGGER.error("UDP Error : " + e.getMessage(), e);
                running = false;
                return;
            }

            byte[] datagram = new byte[indatagram.getLength()];
            System.arraycopy(indatagram.getData(), 0, datagram, 0, indatagram.getLength());
            final CATTPPdu incomingPdu = new CATTPPdu(datagram);
            int dstPort = incomingPdu.getHeader().getDstPort();
            CATTPSession session = getSessionByPort(dstPort);
            receivedPdu = incomingPdu;
            String msisdnStr = "";
            if (session != null) {
                msisdnStr = session.getMsisdn().getValue();
            }

            if (!incomingPdu.verifyChecksum()) {
                LOGGER.error(String.format("[%s] - Wrong checksum, discard", msisdnStr));
                continue;
            }
            if (session == null) {
                // send appropriate reset                
                LOGGER.warn(String.format("%s <BR/>The session is no more ON for port : %s ", incomingPdu, dstPort));
                sendAppropriateReset(indatagram, incomingPdu);
                continue;
            }

            if (!incomingPdu.isEAKSet()) {
                LOGGER.info("--[ UICC ]--->> " + incomingPdu + "<BR/> for " + session);
            } else {
                LOGGER.warn("--[ UICC ]--->> " + incomingPdu + "<BR/> for " + session);
            }


            session.setIpAddress(indatagram.getAddress());
            session.setUdpPort(indatagram.getPort());
            incomingPdu.setSession(session);
            session.setReceivedPdu(incomingPdu);
            session.processIncomingPdu(incomingPdu);
        }
    }

    /**
     * start the server
     */
    public void start() {
        cattpServerThread = new Thread(this);
        cattpServerThread.setDaemon(true);
        cattpServerThread.start();
        LOGGER.debug("CATTP server thread started successfully !");
    }

    /**
     * terminate the session
     */
    public void terminate() {
        Set<Integer> keySet = sessionsByCattpPort.keySet();
        for (Integer sessionKey : keySet) {
            CATTPSession session = (CATTPSession) sessionsByCattpPort.get(sessionKey);
            session.close();
        }
    }

    @SuppressWarnings("unused")
    @PostConstruct
    private void initPublicAccessAndStart() throws MalformedURLException, IOException {
        initPublicAccess();
        LOGGER.info("CATTP server : " + identity + " with public ip address "
                + publicIpAddress + " is started !");
        LOGGER.info("CATTP server retries set to " + retriesCounter);
        start();
    }

    private CATTPSession createNewSession(int cattpPort, Msisdn msisdn,
            String sessionId) throws NoMorePortAvailableException {
        if (cattpPort == 0) {
            throw new NoMorePortAvailableException("No more CATTP port Available, session not created for " + msisdn);
        }

        CATTPSession session = (CATTPSession) sessionLocator.createTransportSession("cattpSession");
        session.setCattpPort(cattpPort);
        session.setMsisdn(msisdn);
        session.setSessionId(sessionId);
        sessionsByCattpPort.put(Integer.valueOf(cattpPort), session);
        return session;
    }

    private void sendAppropriateReset(DatagramPacket inMsg, CATTPPdu pdu) {
        int ack;
        byte reasonCode;

        if (pdu.isACKSet()) {
            ack = pdu.getHeader().getSeq() + 1;
            reasonCode = RstPdu.UNEXPECTED_PDU_RECEIVED;
        } else {
            ack = pdu.getHeader().getSeq();
            if (pdu.isSYNSet()) {
                reasonCode = RstPdu.TEMPORARILY_UNABLE_TO_SET_UP_THIS_CONNECTION;
            } else {
                reasonCode = RstPdu.UNEXPECTED_PDU_RECEIVED;
            }
        }

        CATTPHeader header = new CATTPHeader(
                0,
                ack,
                pdu.getHeader().getDstPort(),
                pdu.getHeader().getSrcPort(),
                1,
                reasonCode);

        RstPdu rstPdu = new RstPdu(header, null);


        CATTPServer.getInstance().sendDataGram(inMsg.getAddress(), inMsg.getPort(),
                rstPdu.getDatagram());
    }

    private int locateAvailablePort() {
        cattpport = FIRST_AVAILABALE_PORT;
        while (cattpport < MAX_AVAILABALE_PORT) {
            if (!sessionsByCattpPort.containsKey(cattpport)) {
                return cattpport;
            }
            cattpport += 2;
        }
        return 0;
    }

    /**
     *
     * @return ConcurrentTaskExecutor
     */
    public ConcurrentTaskExecutor getTaskExecutor() {
        return taskExecutor;
    }

    /**
     *
     * @return CATTPPdu
     */
    public synchronized CATTPPdu getReceivedPdu() {
        return receivedPdu;
    }

    /**
     *
     * @return sessionIdle Life Time
     */
    public int getSessionIdleLifeTime() {
        return sessionIdleLifeTime;
    }

    /**
     *
     * @return identity
     */
    public String getIdentity() {
        return identity;
    }

    /**
     *
     * @return ackPdu Retransmission Timer Value
     */
    public int getAckPduRetransmissionTimerValue() {
        return ackPduRetransmissionTimerValue;
    }

    /**
     *
     * @return nulPdu Retransmission Timer Value
     */
    public int getNulPduRetransmissionTimerValue() {
        return nulPduRetransmissionTimerValue;
    }

    /**
     *
     * @return retries Counter
     */
    public int getRetriesCounter() {
        return retriesCounter;
    }

    /**
     *
     * @return cattpServer Thread
     */
    public Thread getCattpServerThread() {
        return cattpServerThread;
    }

    /**
     * Gets number of living sessions
     *
     * @return the nb living sessions
     */
    public int getNbLivingSessions() {
        return sessionsByCattpPort.size();
    }

    /**
     * Retrieves a Cattp session
     *
     * @param cattpPort search criteria
     * @return returns the session associated to that port
     */
    public CATTPSession getSessionByPort(int cattpPort) {
        return (CATTPSession) sessionsByCattpPort.get(Integer.valueOf(cattpPort));
    }

    /**
     * Retrieves a Cattp session
     *
     * @param msisdn search criteria
     * @return returns the session associated with that msisdn
     * @throws SessionNotFoundException
     */
    public CATTPSession getSessionByMsisdn(Msisdn msisdn) {
        CATTPSession session = (CATTPSession) sessionsByMsisdn.get(msisdn.getValue());
        return session;
    }

    /**
     * Links a msisdn to a session
     *
     * @param msisdn msisdn
     * @param session session
     */
    public void linkSessionToMsisdn(Msisdn msisdn, CATTPSession session) {
        sessionsByMsisdn.put(msisdn.getValue(), session);
    }
}
