/*-------------------------------------------------------- 
 * Module Name : cattp and sms lib
 * Version : 1.7.1
 * 
 * Software Name : Transport Manager
 * Version : 1.7.1
 *
 * Copyright © 01-10-2012 – 01-10-2012 France Télécom
 * This software is distributed under the Apache (v2) license,
 * the text of which is available at http://www.apache.org/licenses/LICENSE-2.0.html
 * or see the "license.txt" file for more details.
 *
 *--------------------------------------------------------
 * File Name   : ${file_name}
 *
 * Created     : ${date}
 * Author(s)   : ${authors}
 *
 * Description :
 * CATTP close-wait state events and transitions
 *
 *--------------------------------------------------------
 * ${Log}
 *
 */
package com.orange.labs.quickpass.tm.cattp.states;

import com.orange.labs.quickpass.tm.cattp.pdu.CATTPPdu;
import com.orange.labs.quickpass.tm.cattp.pdu.RstPdu;
import com.orange.labs.quickpass.tm.cattp.server.CATTPServer;
import com.orange.labs.quickpass.tm.session.CATTPSession;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

/**
 * The CLOSE-WAIT state is entered from either a Close request or from the reception of a RST PDU.<BR/> In that case a CAT_TP
 * entity shall wait for a delay period without initiating any activity on the connection and then enter the CLOSE state.
 */
public final class CATTPCloseWait extends CATTPState {

    private static CATTPCloseWait instance = null;
    private Runnable scheduleClose;

    private CATTPCloseWait() {
    }

    /**
     * @return the instance
     */
    public static CATTPCloseWait getInstance() {
        if (instance == null) {
            instance = new CATTPCloseWait();
        }
        return instance;
    }

    @Override
    public String toString() {
        return " C L O S E - W A I T ";
    }

    //{@inheritDoc} 
    @Override
    public synchronized void entry(final CATTPSession session) {
        // envoi du RESET        
        session.sendRST(RstPdu.NORMAL_ENDING_REASON_CODE);
        scheduleClose = new Runnable() {
            @Override
            public void run() {
                // nouveau RST pour être sûr
                // session.sendRST(RstPdu.NORMAL_ENDING_REASON_CODE);                
                session.setState(CATTPClosed.getInstance());
            }
        };       
        ((ScheduledThreadPoolExecutor) CATTPServer.getInstance()
                .getTaskExecutor().getConcurrentExecutor())
                .schedule(scheduleClose, 15, TimeUnit.SECONDS);
    }

    //{@inheritDoc} 
    @Override
    public void exit(CATTPSession session) {
    }

    //{@inheritDoc} 
    @Override
    public synchronized void receiveACK(CATTPSession session) {
        CATTPPdu currentPdu = session.getReceivedPdu();
        // Send another pdu unless current win size == 0

        if (currentPdu == null) {
            return;
        }

        if (currentPdu.hasData() || currentPdu.isNULSet()) {
            session.setLastInSequencePdu(currentPdu);
            session.sendACK(1, null);
            session.ackPdu(currentPdu.getHeader().getAck());

            if (!session.getOutOfSequencePduList().isEmpty()) {
                CATTPPdu outOfSeqPdu = null;
                do {
                    outOfSeqPdu = session.findOutOfSequencePdu(currentPdu.getHeader().getSeq() + 1);
                    if (outOfSeqPdu != null) {
                        session.storeOutOfSequencePdu(outOfSeqPdu);
                    }
                } while (outOfSeqPdu != null);
            }

            if (currentPdu.hasData()) {
                session.getPduReceivedList().put(currentPdu.getHeader().getSeq(), currentPdu);
                session.onCattpDataReceived(session);
            }
        } else {
            if (!session.ackPdu(currentPdu.getHeader().getAck())) {
                return;
            }
        }
    }
}
