/*-------------------------------------------------------- 
 * Module Name : cattp and sms lib
 * Version : 1.7.1
 * 
 * Software Name : Transport Manager
 * Version : 1.7.1
 *
 * Copyright © 01-10-2012 – 01-10-2012 France Télécom
 * This software is distributed under the Apache (v2) license,
 * the text of which is available at http://www.apache.org/licenses/LICENSE-2.0.html
 * or see the "license.txt" file for more details.
 *
 *--------------------------------------------------------
 * File Name   : ${file_name}
 *
 * Created     : ${date}
 * Author(s)   : ${authors}
 *
 * Description :
 * he OPEN state exists when a connection has been established
 *
 *--------------------------------------------------------
 * ${Log}
 *
 */
package com.orange.labs.quickpass.tm.cattp.states;

import com.orange.labs.quickpass.tm.cattp.pdu.AckPdu;
import com.orange.labs.quickpass.tm.cattp.pdu.CATTPHeader;
import com.orange.labs.quickpass.tm.cattp.pdu.CATTPPdu;
import com.orange.labs.quickpass.tm.session.CATTPSession;
import java.util.SortedMap;

/**
 * The OPEN state exists when a connection has been established. This is the result of a three way handshake: <BR/> • SYN. <BR/>
 * SYN-ACK. <BR/> • ACK. <BR/> between a CAT_TP client and a CAT_TP server. In this state, the two entities are able to begin a
 * full-duplex communication.
 */
public final class CATTPOpen extends CATTPState {

    private static CATTPOpen instance = null;

    private CATTPOpen() {
    }

    /**
     * @return the instance
     */
    public static CATTPOpen getInstance() {
        if (instance == null) {
            instance = new CATTPOpen();
        }
        return instance;
    }

    //{@inheritDoc}
    @Override
    public synchronized void entry(final CATTPSession session) {
        session.onCattpConnectionEstablished(session);
        CATTPPdu pdu = session.getReceivedPdu();

        int windowSize = pdu.getHeader().getWinSize();


        if (pdu.isACKSet()) {
            while (windowSize > 0) {
                CATTPPdu pduToSend = session.getNextPdu();
                if (pduToSend != null) {
                    int nextSeq = pduToSend.getHeader().getSeq();
                    session.setSequenceNumber(nextSeq);
                    session.sendACK((AckPdu) pduToSend);
                }
                windowSize--;
            }
        }

        // inc sequence Number
        session.incSequenceNumber();
    }

    @Override
    public void exit(CATTPSession session) {
    }

    @Override
    public synchronized void processIncomingPdu(CATTPPdu incomingPdu) {
        CATTPSession session = incomingPdu.getSession();

        int rcvCurSeqNb = session.getLastInSequencePdu().getHeader().getSeq();

        if (incomingPdu.isRSTSet()) {
            receiveRST(session);
            return;
        }
        int incommingPduSeq = incomingPdu.getHeader().getSeq();

        if ((rcvCurSeqNb < incommingPduSeq) || (incommingPduSeq <= rcvCurSeqNb)) {
            if (incomingPdu.isSYNSet()) {
                receiveSYN(session);
                return;
            }
            if (incomingPdu.isACKSet()) {
                receiveACK(session);
                return;
            }

            if (incomingPdu.isRSTSet()) {
                receiveRST(session);
            }

        } else {
            session.setReceivedAckablePdu(incomingPdu);
            if (session.getNextPdu() != null) {
                session.sendACK(1, null);
            } else {
                session.sendACK(1, null);
            }
            logger.warn(String.format("[%s]: Discard PDU %s. Expected sequence > %s and <= %s",
                    session, incommingPduSeq, rcvCurSeqNb, rcvCurSeqNb + session.getReceivedPdu().getHeader().getWinSize()));
        }
    }

  
    @Override
    public void closeRequest(CATTPSession session) {
        session.setState(CATTPCloseWait.getInstance());        
    }

    /**
     * Voir 102 127 figure 25
     *
     * @param session CATTPSession
     */
    @Override
    public synchronized void receiveACK(CATTPSession session) {
        CATTPPdu rcvPdu = session.getReceivedPdu();
        // Send another pdu unless current win size == 0
        if (rcvPdu == null) {
            logger.warn(String.format("%s Received pdu null", session));
            return;
        }
        CATTPHeader rcvPduHeader = rcvPdu.getHeader();
        int ack = rcvPduHeader.getAck();
        session.ackPdu(ack);

        if (rcvPdu.hasData() || rcvPduHeader.getWinSize() == 0) {
            session.setLastInSequencePdu(rcvPdu);
            if (!rcvPdu.hasData() && rcvPduHeader.getWinSize() == 0) {
                if (!session.getSentPdu().isNULSet()) {
                    try {
                        // La sim est sous l'eau, on attend 1s;
                        Thread.sleep(1000);
                    } catch (InterruptedException ex) {
                        logger.error(String.format("%s Not possible to interupt Cattp session", session));
                    }
                    // réémission de ACK NUL ? : pas disponible pour l'instant
                    session.sendNUL();
                    session.incSequenceNumber();
                }
            } else {
                session.sendACK(1, null);
            }
            int rcvPduHeaderSeq = rcvPduHeader.getSeq();
            if (!session.getOutOfSequencePduList().isEmpty()) {
                CATTPPdu outOfSeqPdu;
                do {
                    outOfSeqPdu = session.findOutOfSequencePdu(rcvPduHeaderSeq + 1);
                    if (outOfSeqPdu != null) {
                        session.storeOutOfSequencePdu(outOfSeqPdu);
                    }
                } while (outOfSeqPdu != null);
            }
            if (rcvPdu.hasData()) {
                session.getPduReceivedList().put(rcvPduHeaderSeq, rcvPdu);
                session.onCattpDataReceived(session);
            }
        } else if (!session.ackPdu(ack) && !(rcvPdu.isNULSet())) {
            return;
        }

        //@FIXME manage eack and retries
        /*
         * if (rcvPdu.isEAKSet()) { session.setSequenceNumber(rcvPdu.getHeader().getAck() + 1); }
         */
        int windowSize = rcvPduHeader.getWinSize();
        int initWinSize = windowSize;

        while (windowSize > 0) {
            int sequenceNumber = session.getSequenceNumber();
            if (initWinSize + ack >= sequenceNumber) {
                SortedMap<Integer, CATTPPdu> pdusInCurrentSdu = session.getPdusInCurrentSdu();
                logger.trace(String.format("%s pduInCurrentSdu seqs: %s - current seq number: %s.",
                        session, pdusInCurrentSdu.keySet(), sequenceNumber));
                CATTPPdu currentSentPdu = session.getPdu(sequenceNumber);
                if (ack == session.getLastSentAckablePdu().getHeader().getSeq() && pdusInCurrentSdu.isEmpty()) {
                    logger.debug(String.format("%s No more PDU for this SDU was found for sequence : %s.",
                            session, sequenceNumber));
                    session.onCattpNoMorePduForCurrentSduToSend(session);
                    //break;
                }

                if (currentSentPdu != null) {
                    session.incSequenceNumber();
                    session.sendACK((AckPdu) currentSentPdu);
                    currentSentPdu.setAck(session.getReceivedAckablePdu().getHeader().getSeq());
                }
            }
            windowSize--;
        }
    }

    @Override
    public synchronized void receiveSYN(CATTPSession session) {
        session.setState(CATTPClosed.getInstance());
    }

    @Override
    public String toString() {
        return " O P E N ";
    }
}
