/*-------------------------------------------------------- 
* Module Name : cattp and sms lib
* Version : 1.7.1
* 
* Software Name : Transport Manager
* Version : 1.7.1
*
* Copyright © 01-10-2012 – 01-10-2012 France Télécom
* This software is distributed under the Apache (v2) license,
* the text of which is available at http://www.apache.org/licenses/LICENSE-2.0.html
* or see the "license.txt" file for more details.
*
*--------------------------------------------------------
* File Name   : ${file_name}
*
* Created     : ${date}
* Author(s)   : ${authors}
*
* Description :
* state machine abstract class for cattp
*
*--------------------------------------------------------
* ${Log}
*
*/

package com.orange.labs.quickpass.tm.cattp.states;

import com.orange.labs.quickpass.tm.cattp.pdu.CATTPPdu;
import com.orange.labs.quickpass.tm.cattp.pdu.RstPdu;
import com.orange.labs.quickpass.tm.session.CATTPSession;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A CAT_TP connection progresses into several states during its lifetime. <br/>
 * The states are shown in the figure 8 ETSI 102 127. <br/>Figure
 * boxes represent CAT_TP states and arrows between them represent state change direction. <br/>
 * The event that triggers the state change is described above the line and the 
 * resulting output is described in figure 8 ETSI 102 127.
 */
public abstract class CATTPState {

    /** LOGGER */
    protected final Logger logger = LoggerFactory.getLogger(CATTPState.class);

    /**
     * action performed when entering the state
     *
     * @param session curent session
     */
    public abstract void entry(CATTPSession session);

    /**
     * action performed when exiting the state
     *
     * @param session curent session
     */
    public abstract void exit(CATTPSession session);

    /**
     * Default process PDU behaviour.
     *
     * @param pdu Current pdu
     */
    public void processIncomingPdu(CATTPPdu pdu) {

        if (pdu.isSYNSet()) {
            receiveSYN(pdu.getSession());
            return;
        }
        if (pdu.isACKSet()) {
            receiveACK(pdu.getSession());
            return;
        }
        if (pdu.isRSTSet()) {
            receiveRST(pdu.getSession());
            return;
        }
    }

    /**
     * Receive SYN PDU event
     *
     * @param session curent session
     */
    public void receiveSYN(CATTPSession session) {
        logger.warn(String.format("%s Unexpected SYN received.", session));
    }

    
    /**
     * Receive ACK PDU event
     *
     * @param session curent session
     */
    public void receiveACK(CATTPSession session) {
        logger.warn(String.format("%s Unexpected ACK received.", session));
        session.sendRST(RstPdu.UNEXPECTED_PDU_RECEIVED);
    }

    
    /**
     * Receive RST PDU event
     *
     * @param session curent session
     */
    public synchronized void receiveRST(CATTPSession session) {
        session.setState(CATTPCloseWait.getInstance());
    }

    
    /**
     * Receive SYN ACK PDU.
     *
     * @param session curent session
     */
    public void receiveSYNACK(CATTPSession session) {
        logger.warn(String.format("%s Unexpected passiveOpen request received.", session));
    }

    
    /**
     * Close request.
     *
     * @param session curent session
     */
    public void closeRequest(CATTPSession session) {
        logger.warn(String.format("%s Unexpected close request received.", session));
    }
}
