/*-------------------------------------------------------- 
 * Module Name : cattp and sms lib
 * Version : 1.7.1
 * 
 * Software Name : Transport Manager
 * Version : 1.7.1
 *
 * Copyright © 01-10-2012 – 01-10-2012 France Télécom
 * This software is distributed under the Apache (v2) license,
 * the text of which is available at http://www.apache.org/licenses/LICENSE-2.0.html
 * or see the "license.txt" file for more details.
 *
 *--------------------------------------------------------
 * File Name   : ${file_name}
 *
 * Created     : ${date}
 * Author(s)   : ${authors}
 *
 * Description :
 * CATTP Session 
 *
 *--------------------------------------------------------
 * ${Log}
 *
 */
package com.orange.labs.quickpass.tm.session;

import com.orange.labs.quickpass.common.utils.ConvertUtils;
import com.orange.labs.quickpass.tm.cattp.CATTPActivityListener;
import com.orange.labs.quickpass.tm.cattp.pdu.AckPdu;
import com.orange.labs.quickpass.tm.cattp.pdu.CATTPHeader;
import com.orange.labs.quickpass.tm.cattp.pdu.CATTPPdu;
import com.orange.labs.quickpass.tm.cattp.pdu.NulPdu;
import com.orange.labs.quickpass.tm.cattp.pdu.RstPdu;
import com.orange.labs.quickpass.tm.cattp.pdu.SynAckPdu;
import com.orange.labs.quickpass.tm.cattp.pdu.SynExtraData;
import com.orange.labs.quickpass.tm.cattp.server.CATTPServer;
import com.orange.labs.quickpass.tm.cattp.states.CATTPClosed;
import com.orange.labs.quickpass.tm.cattp.states.CATTPListen;
import com.orange.labs.quickpass.tm.cattp.states.CATTPState;
import java.net.InetAddress;
import java.util.HashMap;
import java.util.Map;
import java.util.Random;
import java.util.SortedMap;
import java.util.TreeMap;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Manages one CATTP Session
 *
 * @author Sylvain Decourval
 *
 */
public class CATTPSession extends AbstractTransportSession implements Runnable, CATTPActivityListener {

    private static final Logger LOGGER = LoggerFactory.getLogger(CATTPSession.class);
    /**
     * Set to min window size
     */
    short sentWindowSize = 1;
    int cattpPort;
    int sequenceNumber = 100;
    private int nbSecurePacketsSent;
    /**
     * Last pdu sent
     */
    private CATTPPdu lastSentAckablePdu = null;
    /**
     * Last pdu sent
     */
    private CATTPPdu sentPdu = null;
    /**
     * Last 'ackable' pdu
     */
    private CATTPPdu receivedAckablePdu = null;
    /**
     * Last in-sequence Pdu
     */
    private CATTPPdu lastInSequencePdu = null;
    /**
     * Last pdu received
     */
    private CATTPPdu receivedPdu = null;
    /**
     * Utiliser si W > 1
     */
    private SortedMap<Integer, CATTPPdu> pdusInCurrentSdu = new TreeMap<Integer, CATTPPdu>();
    //private SortedMap<Integer, CATTPPdu> pdusInCurrentSdu = Collections.synchronizedSortedMap(pduList);
    private Map<Integer, CATTPPdu> outOfSequencePduList = new HashMap<Integer, CATTPPdu>();
    private SortedMap<Integer, CATTPPdu> pduReceivedList = new TreeMap<Integer, CATTPPdu>();
    private InetAddress ipAddress;
    private int udpPort;
    private CATTPState state = null;
    private short maxPduSize;
    private short maxSduSize;
    /**
     * Identifier received during SYNC phase
     */
    private byte[] ident;
    /**
     * Initialy received sequence number
     */
    private int rcvIniSeqNb;
    /**
     * initially received window size
     */
    private int rcvIniWinSize;
    private int pdusIndex = 0;
    private byte resetCause = RstPdu.NORMAL_ENDING_REASON_CODE;

    /**
     * Cattp Session constructor
     */
    public CATTPSession() {
        Random rnd = new Random();
        this.sequenceNumber = 100 + rnd.nextInt(10) * 100;
        LOGGER.debug(String.format("This session initial sequence number is : %s", sequenceNumber));
        state = CATTPClosed.getInstance();
    }

    /**
     * Getter for PDUs out of sequence
     * @return map of PDUs out of sequence
     */
    public Map<Integer, CATTPPdu> getOutOfSequencePduList() {
        return outOfSequencePduList;
    }

    /**
     * close cattp session
     */
    public void close() {
        LOGGER.info("Closing CATTP Session:" + this.toString());
        state.closeRequest(this);
    }

    /**
     * Change the CATTP state. This function calls succesively the exit method of the current state and the entry method of the
     * next state
     *
     * @param state next state
     */
    public void setState(CATTPState state) {
//        if (this.state == state) {
//            return;
//        }
        LOGGER.info(String.format("%s # CATTP / State {PREVIOUS = [%s] -> CURRENT=[%s] #}", this, this.state, state));
        if (this.state != null) {
            this.state.exit(this);
        }
        this.state = state;
        state.entry(this);
    }

    /**
     * Turns secure packets to Cattp sdu
     */
    public void provisionPackets() {
        provisionPackets(sequenceNumber);
    }

    /**
     * Turns secure packets to Cattp sdu
     *
     * @param sequenceNumber sequenceNumber
     */
    public void provisionPackets(int sequenceNumber) {
        CATTPPdu pdu;
        if ((securePackets == null) || (securePackets.isEmpty())) {
            return;
        }

        if (securePackets != null) {
            nbSecurePacketsSent = securePackets.size();
            byte[] securePacket = securePackets.get(0);

            String[] pdusData = ConvertUtils.splitHexString(ConvertUtils.toHexString(securePacket),
                    this.maxPduSize - 18);
            boolean seg = true;
            for (String pduData : pdusData) {
                if (pduData.equals(pdusData[pdusData.length - 1])) {
                    seg = false;
                }

                CATTPHeader header = new CATTPHeader(++sequenceNumber, receivedAckablePdu.getHeader().getSeq(),
                        receivedPdu.getHeader().getDstPort(), receivedPdu.getHeader().getSrcPort(), 1);

                pdu = new AckPdu(header, ConvertUtils.hex2byte(pduData), seg, this);

                pdusInCurrentSdu.put(pdu.getHeader().getSeq(), pdu);
            }
            LOGGER.debug(String.format("%s There are %s secured packets and %s "
                    + "pdus for the current sdus to process.\n",
                    this, securePackets.size(), pdusInCurrentSdu.size()));
            securePackets.remove(0);
            pdusIndex = pdusInCurrentSdu.firstKey();
        }
    }

    /**
     * Saves an out of sequence pdu
     *
     * @param pdu out of sequence pdu
     */
    public void flushOutOfSequencePdu(CATTPPdu pdu) {
        if (pdu != null) {
            outOfSequencePduList.put(pdu.getHeader().getSeq(), pdu);
        }
    }

    /**
     * Finds an out of sequence pdu in the ot of seq pdus list
     *
     * @param seq sequence to find
     * @return out of sequence CATTPdu
     */
    public CATTPPdu findOutOfSequencePdu(int seq) {
        return outOfSequencePduList.get(seq);
    }

    /**
     * Stores out of sequence PDU 
     *
     * @param pdu out of sequence CATTPdu
     */
    public void storeOutOfSequencePdu(CATTPPdu pdu) {
        pdusInCurrentSdu.put(pdu.getHeader().getSeq(), pdu);
        removeOutOfSequencePdu(pdu.getHeader().getSeq());
    }

    /**
     * removes an out of sequence pdu
     *
     * @param seq Pdu seq number
     */
    public void removeOutOfSequencePdu(int seq) {
        outOfSequencePduList.remove(seq);
    }

    /**
     * Sends Ack pdu
     */
    public void sendSYNACK() {
        SynExtraData synData = new SynExtraData(getMaxPduSize(), getMaxSduSize(),
                CATTPServer.getInstance().getIdentity().getBytes());
        CATTPHeader rPduHeader = receivedPdu.getHeader();
        CATTPHeader header = new CATTPHeader(sequenceNumber, rPduHeader.getSeq(), rPduHeader.getDstPort(),
                rPduHeader.getSrcPort(), getSentWindowSize(), synData);

        sentPdu = new SynAckPdu(header, this);
        lastSentAckablePdu = sentPdu;
        sentPdu.sendAndScheduleNextRetry();
    }

    /**
     * Sends RST PDU
     */
    public void sendRST() {
        CATTPHeader header = new CATTPHeader(sequenceNumber, receivedAckablePdu.getHeader().getSeq(),
                receivedPdu.getHeader().getDstPort(), receivedPdu.getHeader().getSrcPort(), getSentWindowSize(),
                resetCause);

        sentPdu = new RstPdu(header, this);
        sentPdu.send();
    }

    /**
     * Sends RST PDU
     *
     * @param resetCause cause of resest
     */
    public void sendRST(byte resetCause) {
        CATTPHeader header = new CATTPHeader(sequenceNumber, receivedAckablePdu.getHeader().getSeq(),
                receivedPdu.getHeader().getDstPort(), receivedPdu.getHeader().getSrcPort(), getSentWindowSize(),
                resetCause);

        sentPdu = new RstPdu(header, this);
        sentPdu.send();
    }

    /**
     * Sends NUL PDU
     */
    public void sendNUL() {
        CATTPHeader header = new CATTPHeader(sequenceNumber, receivedAckablePdu.getHeader().getSeq(),
                receivedPdu.getHeader().getDstPort(), receivedPdu.getHeader().getSrcPort(), getSentWindowSize());

        sentPdu = new NulPdu(header, this);
        lastSentAckablePdu = sentPdu;
        //@Fixme retries sur les NUL ne s'arreteent visiblement pas
        //sentPdu.sendAndScheduleNextRetry();
        sentPdu.send();
    }

    /**
     * Sends ACK pdu
     *
     * @param win window size
     * @param securePacket if null no data sent
     */
    public void sendACK(int win, byte[] securePacket) {
        CATTPHeader header = new CATTPHeader(sequenceNumber, receivedAckablePdu.getHeader().getSeq(),
                receivedPdu.getHeader().getDstPort(), receivedPdu.getHeader().getSrcPort(), win);

        sentPdu = new AckPdu(header, securePacket, false, this);

        if (securePacket != null) {
            sentPdu.sendAndScheduleNextRetry();
            lastSentAckablePdu = sentPdu;
        } else {
            sentPdu.send();
        }
    }

    /**
     * Sends ACK pdu
     *
     * @param ackPdu ackPdu
     */
    public void sendACK(AckPdu ackPdu) {
        ackPdu.setAck(receivedAckablePdu.getHeader().getSeq());
        sentPdu = ackPdu;
        if (ackPdu.hasData()) {
            sentPdu.sendAndScheduleNextRetry();
            lastSentAckablePdu = sentPdu;
        } else {
            sentPdu.send();
        }
    }

    /**
     * Increments this session seq num
     */
    public void incSequenceNumber() {
        sequenceNumber++;
    }

    /**
     * Process Incoming PDU (from CATTP server)
     *
     * @param pdu pdu
     */
    public void processIncomingPdu(CATTPPdu pdu) {
        state.processIncomingPdu(pdu);
    }

    @Override
    public void onCattpDataReceived(CATTPSession session) {
        CATTPPdu receivedPduInSession = session.getReceivedPdu();
        session.ackPdu(receivedPduInSession.getHeader().getAck());
        //session.getReceivedPdu().cancelRetry();
        LOGGER.debug(String.format("%s Received data %s/%s for session:%s", session,
                ConvertUtils.toHexString(receivedPduInSession.getData(), ":", 40),
                nbSecurePacketsSent, Integer.toHexString(session.getCattpPort())));
        LOGGER.debug(String.format("%s Pdus left to send : %s", session, pdusInCurrentSdu.size()));


        if (transportReply != null) {
            // @FIXME : manage several PDUs in response (see revision befor 17494)
            transportReply.handleSEResponse(session.getMsisdn(), receivedPduInSession.getData());
        } else {
            LOGGER.warn(String.format("%s Transport reply service not set ! Please initialize it correctly !", session));
        }

        pduReceivedList.clear();
        //nbSecurePacketsSent = 0;

    }

    @Override
    public void run() {
        // Passive Session begins in Listen State
        setState(CATTPListen.getInstance());
    }

    //{@inheritDoc} 
    @Override
    public void onCattpConnectionEstablished(CATTPSession session) {
        ackPdu(session.getReceivedPdu().getHeader().getAck());
        //session.getReceivedPdu().cancelRetry();
    }

    //{@inheritDoc} 
    @Override
    public void onCattpSDUSuccessfullyReceived(CATTPSession session) {
        LOGGER.info(String.format("%s  SDU  has been received :)", session));
    }

    //{@inheritDoc} 
    @Override
    public void onStatus(CATTPSession session) {
        LOGGER.info(String.format("%s  onStatus raised.", session));
    }

    /**
     * Used to ack a sent pdu
     *
     * @param sequenceNumber sequenceNumber
     * @return true or false
     */
    public boolean ackPdu(int sequenceNumber) {

        for (int i = 0; i < rcvIniWinSize; i++) {
            CATTPPdu pdu = (CATTPPdu) pdusInCurrentSdu.get(sequenceNumber - i);
            removeAcknowledgedPdu(sequenceNumber - i);
            if (pdu != null) {
                pdu.cancelRetry();
            }
        }
        return true;
    }

    /**
     * Next pdu in the SDU
     *
     * @return next CATTPPdu or null if last PDU already get
     */
    public synchronized CATTPPdu getNextPdu() {
        if (pdusInCurrentSdu.isEmpty()) {
            return null;
        }
        return (CATTPPdu) pdusInCurrentSdu.get(pdusIndex++);
    }

    /**
     * Getter Pdu 
     * @param index index
     * @return cATTPPdu
     */
    public CATTPPdu getPdu(int index) {
        pdusIndex = index;
        return getNextPdu();
    }

    @Override
    public void onCattpNoMorePduForCurrentSduToSend(CATTPSession session) {
        provisionPackets(lastSentAckablePdu.getHeader().getSeq());
        if (pdusInCurrentSdu.isEmpty()) {
            LOGGER.debug(String.format("%s All SDU have been sent (onCattpNoMorePduForCurrentSduToSend)", session));
        } else {
            LOGGER.debug(String.format("%s has " + securePackets.size() + " more SDU(s) to send.", session));
        }
    }

    @Override
    public String toString() {
        return String.format("CATTP session[%s/%s:%s]<br/>", msisdn, sessionId, Integer.toHexString(cattpPort));
    }

    /**
     * Getter for the last PDU in sequence
     * @return PDU
     */
    public CATTPPdu getLastInSequencePdu() {
        return lastInSequencePdu;
    }

    /**
     * setter for the last PDU in sequence
     * @param lastInSequencePdu
     */
    public void setLastInSequencePdu(CATTPPdu lastInSequencePdu) {
        LOGGER.debug(String.format("%s Last in-sequence pdu received : %s.", this, lastInSequencePdu.getHeader().getSeq()));
        this.lastInSequencePdu = lastInSequencePdu;
    }

    /**
     * getter for the list of received PDUs
     * @return list of received PDUs
     */
    public SortedMap<Integer, CATTPPdu> getPduReceivedList() {
        return pduReceivedList;
    }

    private void removeAcknowledgedPdu(int sequenceNumber) {
        pdusInCurrentSdu.remove(sequenceNumber);
    }

    /**
     * getter cattp state associated to the current version
     * @return CATTP state
     */
    public CATTPState getState() {
        return state;
    }

    /**
     * getter IP address
     * @return current ip address
     */
    public InetAddress getIpAddress() {
        return ipAddress;
    }

    /**
     * setter IP adress
     * @param ipAddress
     */
    public void setIpAddress(InetAddress ipAddress) {
        this.ipAddress = ipAddress;
    }

    /**
     * getter udp port
     * @return udp port
     */
    public int getUdpPort() {
        return udpPort;
    }

    /**
     * setter udp port
     * @param udpPort udp port
     */
    public void setUdpPort(int udpPort) {
        this.udpPort = udpPort;
    }

    /**
     * getter received PDU
     * @return received PDU
     */
    public CATTPPdu getReceivedPdu() {
        return receivedPdu;
    }

    /**
     * setter received Pdu
     * @param receivedPdu received PDU
     */
    public void setReceivedPdu(CATTPPdu receivedPdu) {
        this.receivedPdu = receivedPdu;
        if (receivedPdu.needsAck()) {
            receivedAckablePdu = receivedPdu;
        }
    }

    /**
     * getter last received ackable PDU
     * @return last received ackable PDU
     */
    public CATTPPdu getReceivedAckablePdu() {
        return receivedAckablePdu;
    }

    /**
     * setter last received ackable PDU
     * @param receivedAckablePdu last received ackable PDU
     */
    public void setReceivedAckablePdu(CATTPPdu receivedAckablePdu) {
        this.receivedAckablePdu = receivedAckablePdu;
    }

    /**
     * getter last sent PDU
     * @return last sent PDU
     */
    public CATTPPdu getSentPdu() {
        return sentPdu;
    }

    /**
     * setter last sent PDU
     * @param sentPdu last sent PDU
     */
    public void setSentPdu(CATTPPdu sentPdu) {
        this.sentPdu = sentPdu;
    }

    /**
     * getter sent windows size
     * @return sent windows size
     */
    public short getSentWindowSize() {
        return sentWindowSize;
    }

    /**
     * setter sent windows size
     * @param sentWindowSize sent windows size
     */
    public void setSentWindowSize(short sentWindowSize) {
        this.sentWindowSize = sentWindowSize;
    }

    /**
     * getter for the current sequence number
     * @return current sequence number
     */
    public int getSequenceNumber() {
        return sequenceNumber;
    }

    /**
     * setter for the current sequence number
     * @param sequenceNumber sequence number
     */
    public void setSequenceNumber(int sequenceNumber) {
        this.sequenceNumber = sequenceNumber;
    }

    /**
     * getter for the PDU size
     * @return PDU size
     */
    public int getMaxPduSize() {
        return maxPduSize;
    }

    /**
     * setter for PDU size
     * @param maxPduSize PDU size
     */
    public void setMaxPduSize(short maxPduSize) {
        this.maxPduSize = maxPduSize;
    }

    /**
     * getter for the SDU size
     * @return SDU size
     */
    public int getMaxSduSize() {
        return maxSduSize;
    }

    /**
     * setter for SDU size
     * @param maxSduSize SDU size
     */
    public void setMaxSduSize(short maxSduSize) {
        this.maxSduSize = maxSduSize;
    }
   

    /**
     * setter for the identification data
     * @param ident identification data
     */
    public void setIdent(byte[] ident) {
        this.ident = ident.clone();
    }
    

    /**
     * setter for the received seq number in the SYN 
     * @param rcvIniSeqNb seq number in the SYN
     */
    public void setRcvIniSeqNb(int rcvIniSeqNb) {
        this.rcvIniSeqNb = rcvIniSeqNb;
    }
  

    /**
     * setter for the received windows size in the SYN
     * @param rcvIniWinSize received windows size
     */
    public void setRcvIniWinSize(int rcvIniWinSize) {
        this.rcvIniWinSize = rcvIniWinSize;
    }

    /**
     * getter for the current cattp port
     * @return cattp port
     */
    public int getCattpPort() {
        return cattpPort;
    }

    /**
     * setter for the current cattp port
     * @param cattpPort cattp port
     */
    public void setCattpPort(int cattpPort) {
        this.cattpPort = cattpPort;
    }

    /**
     * getter for the list od PDUs in current SDU
     * @return list of PDUs
     */
    public SortedMap<Integer, CATTPPdu> getPdusInCurrentSdu() {
        return pdusInCurrentSdu;
    }

    /**
     * get the last sent ackable PDU
     * @return last sent ackable PDU
     */
    public CATTPPdu getLastSentAckablePdu() {
        return lastSentAckablePdu;
    }
    

    /**
     * setter for the reset cause
     * @param resetCause
     */
    public void setResetCause(byte resetCause) {
        this.resetCause = resetCause;
    }
}
